from fvcore.nn import FlopCountAnalysis
import math
import sys
from typing import Iterable
import torch
import torch.nn as nn
import utils
from einops import rearrange
from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
import numpy as np

def train_one_epoch(args, model: torch.nn.Module, data_loader: Iterable, optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler, max_norm: float = 0, patch_size: int = 16, 
                    normlize_target: bool = True, log_writer=None, lr_scheduler=None, start_steps=None,
                    lr_schedule_values=None, wd_schedule_values=None):
    model.train()
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    metric_logger.add_meter('min_lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 10
    if args.mcs_loss_based:
        loss_func = nn.MSELoss(reduction='none')
    else:
        loss_func = nn.MSELoss()
    for step, batch in enumerate(metric_logger.log_every(data_loader, print_freq, header)):
        # assign learning rate & weight decay for each step
        it = start_steps + step  # global training iteration
        if lr_schedule_values is not None or wd_schedule_values is not None:
            for i, param_group in enumerate(optimizer.param_groups):
                if lr_schedule_values is not None:
                    param_group["lr"] = lr_schedule_values[it] * param_group["lr_scale"]
                if wd_schedule_values is not None and param_group["weight_decay"] > 0:
                    param_group["weight_decay"] = wd_schedule_values[it]

        videos, bool_masked_pos, indices = batch
        videos = videos.to(device, non_blocking=True)
        if args.mask_type != 'mcm':
            bool_masked_pos = bool_masked_pos.to(device, non_blocking=True).flatten(1).to(torch.bool)

        with torch.no_grad():
            if args.mcs:
                if args.distributed:
                    patch_embed = model.module.encoder.patch_embed
                else:
                    patch_embed = model.encoder.patch_embed                    

                ids = motion_centric_sampling(patch_embed, videos, args.mcm_ratio)
                if args.mcm_vis:
                    origin_videos = videos.clone()
                videos = torch.gather(videos, dim=2, index=ids.repeat(1,3,1,224,224))
            # calculate the predict label
            mean = torch.as_tensor(IMAGENET_DEFAULT_MEAN).to(device)[None, :, None, None, None]
            std = torch.as_tensor(IMAGENET_DEFAULT_STD).to(device)[None, :, None, None, None]
            unnorm_videos = videos * std + mean  # in [0, 1]

            if normlize_target:
                videos_squeeze = rearrange(unnorm_videos, 'b c (t p0) (h p1) (w p2) -> b (t h w) (p0 p1 p2) c', p0=2, p1=patch_size, p2=patch_size)
                videos_norm = (videos_squeeze - videos_squeeze.mean(dim=-2, keepdim=True)
                    ) / (videos_squeeze.var(dim=-2, unbiased=True, keepdim=True).sqrt() + 1e-6)
                # we find that the mean is about 0.48 and standard deviation is about 0.08.
                videos_patch = rearrange(videos_norm, 'b n p c -> b n (p c)')
            else:
                videos_patch = rearrange(unnorm_videos, 'b c (t p0) (h p1) (w p2) -> b (t h w) (p0 p1 p2 c)', p0=2, p1=patch_size, p2=patch_size)

            B, _, C = videos_patch.shape
            if args.mask_type != 'mcm':
                labels = videos_patch[bool_masked_pos].reshape(B, -1, C)

        with torch.cuda.amp.autocast():
            outputs, (_, mcm_target_mask) = model(videos, bool_masked_pos)
            if args.mask_type == 'mcm':
                labels = videos_patch[~mcm_target_mask].reshape(B, -1, C)
            loss = loss_func(input=outputs, target=labels)
            if args.mcs_loss_based:
                for i, index in enumerate(indices):
                    data_loader.dataset.loss_based_windows[index.item()] = int((args.num_frames * args.mcs_ratio - args.num_frames) * math.exp(-1*loss[i].mean().item()) + args.num_frames)
                loss = loss.mean()
        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        optimizer.zero_grad()
        # this attribute is added by timm on one optimizer (adahessian)
        is_second_order = hasattr(optimizer, 'is_second_order') and optimizer.is_second_order
        grad_norm = loss_scaler(loss, optimizer, clip_grad=max_norm,
                                parameters=model.parameters(), create_graph=is_second_order)
        loss_scale_value = loss_scaler.state_dict()["scale"]

        torch.cuda.synchronize()

        metric_logger.update(loss=loss_value)
        metric_logger.update(loss_scale=loss_scale_value)
        min_lr = 10.
        max_lr = 0.
        for group in optimizer.param_groups:
            min_lr = min(min_lr, group["lr"])
            max_lr = max(max_lr, group["lr"])

        metric_logger.update(lr=max_lr)
        metric_logger.update(min_lr=min_lr)
        weight_decay_value = None
        for group in optimizer.param_groups:
            if group["weight_decay"] > 0:
                weight_decay_value = group["weight_decay"]
        metric_logger.update(weight_decay=weight_decay_value)
        metric_logger.update(grad_norm=grad_norm)

        if log_writer is not None:
            log_writer.update(loss=loss_value, head="loss")
            log_writer.update(loss_scale=loss_scale_value, head="opt")
            log_writer.update(lr=max_lr, head="opt")
            log_writer.update(min_lr=min_lr, head="opt")
            log_writer.update(weight_decay=weight_decay_value, head="opt")
            log_writer.update(grad_norm=grad_norm, head="opt")
            log_writer.set_step()

        if lr_scheduler is not None:
            lr_scheduler.step_update(start_steps + step)
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}

def motion_centric_sampling(patch_embed, videos, mcm_ratio):
    T = videos.shape[2] // 2
    x = patch_embed(videos)
    N, L, D = x.shape  # batch, length, dim
    H, W = int(math.sqrt(L / T)), int(math.sqrt(L / T))
    patch_embed_vectors = x.reshape(shape=(N, T, H*W, D))
    
    dist = torch.norm(patch_embed_vectors[:,0,:,:] - patch_embed_vectors[:,1,:,:], p=2, dim=2)
    dist_list = [dist]
    for i in range(T-2):
        dist_list.append(torch.norm(patch_embed_vectors[:,i+1,:,:] - patch_embed_vectors[:,i+2,:,:], p=2, dim=2))
    distance = torch.cat(dist_list, dim=1)
    
    # distance = torch.norm(patch_embed_vectors[:,0,:,:] - patch_embed_vectors[:,1,:,:], p=2, dim=2)
    # for i in range(T-2):
    #     distance = torch.cat((distance, torch.norm(patch_embed_vectors[:,i+1,:,:] - patch_embed_vectors[:,i+2,:,:], p=2, dim=2)), dim=1)
    
    sort_idx = torch.argsort(distance, dim=1, descending=True)
    num_compressed_tokens = int((1 - mcm_ratio) * (T * H * W))
    sort_idx = sort_idx[:,:num_compressed_tokens]
    

    importance = torch.zeros(N, (T-1)*H*W , device=x.device).scatter_(1, sort_idx, 1)
    importance = importance.reshape(N, T-1, H*W).sum(2).cpu() + 1e-2
    ids = []
    
    for i in range(N):
        sampled_id = np.random.choice(
            len(importance[i]), 
            16//2 - 1, 
            replace=False, 
            p=np.array(importance[i]) / (1e-2 * len(importance[i]) + num_compressed_tokens)
            ) + 1
        sampled_id = np.insert(sampled_id, 0, 0) * 2
        sampled_id = np.append(sampled_id, sampled_id + 1)
        sampled_id.sort()
        ids.append(sampled_id)
    
    ids = torch.tensor(ids).cuda(non_blocking=True).unsqueeze(1).unsqueeze(-1).unsqueeze(-1)                
    return ids